# bash-csv: functions for handling CSV files in bash
# Copyright (C)  2019 Heinz Deinhart
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# bash-csv version 0.1
# RFC: https://tools.ietf.org/html/rfc4180

#@@ error messages

#@ error-codes (bash-csv)
# <50 .. reserved for base-base error-codes
#  51 .. illegal handle-name
#  52 .. handle not loaded
#  53 .. line not found
#  54 .. row not found
#  55 .. handle already in use
#  56 .. parse error
#  9x .. errors local to each function

#@ bcsv_string_of_error_code <ret-vn> <error-code>
# returns string representation of error code
# errors: 2, 3
function bcsv_error_string() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _err_code="$2"

    bb_validate_vn bb_error_string "$_ret_vn" || return 3

    if (( "$_err_code" < 50 ))
    then
	bb_error_string "$_ret_vn" "$_err_code"
    else
	local _msg

	case "$_err_code" in
	    51) _msg="illegal db-name";;
	    52) _msg="db named db-name not loaded";;
	    53) _msg="section not found";;
	    54) _msg="key not found";;
	    55) _msg="db named db-name already loaded";;
	    56) _msg="parse error";;
	    90|91|92|93|94|95|96|97|98|99) _msg="function local error";;
	    *) _msg="bb_error_string: unknown error code $_err_code";;
	esac
	printf -v "$_ret_vn" "%s" "$_msg"
    fi
}

#@@ Utils

#@ csv_delimiter_validate <delimiter>
function bcsv_delimiter_validate() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _deli="$1"

    (( ${#_deli} == 1 )) || return 91
}

# _bcsv_escaped_delimiter <ret-vn> <delimiter>
function bcsv_regex_of_delimiter() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    bb_validate_vn "ret" "$_ret_vn"

    local _ret_vn="$1"
    local _deli="$2"

    if [[ "$_deli" =~ ^[\\^$.|?*+(){[]$ ]]
    then
	printf -v "$_ret_vn" '[\%s]' "$_deli"
    else
	printf -v "$_ret_vn" '[%s]' "$_deli"
    fi
}

#@ bcsv_quote_field <ret-vn> <field> <delimiter>
# returns field, quoted if it contains delimiter characters or quotation marks
function bcsv_quote_field() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _field="$2"
    local _deli="$3"

    bb_validate_vn "ret" "$_ret_vn" || return 3

    # shellcheck disable=SC2034
    local bcsv_quote_field__deli_regex
    bcsv_regex_of_delimiter bcsv_quote_field__deli_regex "$_deli" || return "$?"

    if [[ "$_field" =~ ${bcsv_quote_field__deli_regex} || "$_field" =~ \" ]]
    then # escaped
	printf -v "$_ret_vn" '"%s"' "${_field//\"/\"\"}"
    else # non-escaped
	printf -v "$_ret_vn" "%s" "$_field"
    fi
}

# _bcsv_fields_of_line <ret-fields-row-array-vn> <delimiter> <csv-line>
#                      [<ret-error-hint-vn>]
# parses a csv style line and returns fields in ret-fields-row-array
# on parse error it will return the parts parsed so far and store
# info about parse error into ret-error-hint if error is 56
# errors: 2, 3, 56 (parse error), 91 (delimiter is not one character)
function _bcsv_fields_of_line() {
    (( $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    local _ret_array_vn="$1"
    local _deli="$2"
    local _line="$3"
    local _err_hint_vn="${4-}"

    bb_validate_vn "_bcsv_fields_of_line:ret-array" "$_ret_array_vn" || return 3

    bcsv_delimiter_validate "$_deli" || return "$?"

    # clear output array
    eval "$_ret_array_vn=()"

    if [[ "$_deli" =~ ^[\\^$.|?*+(){[]$ ]]
    then
	printf -v _deli '\%s' "$_deli"
    fi

    local _string

    _part="$_line"
    while (( ${#_part} > 0 ))
    do
	if [ "${_part:0:1}" == '"' ]
	then
	    # escaped
	    if [[ "$_part" =~ ^[\"]((\"{2}|[^\"])*)[\"](${_deli}?)(.*)$ ]]
	    then
		# shellcheck disable=SC2034
		_string="${BASH_REMATCH[1]//\"\"/\"}"
		eval "$_ret_array_vn+=(\"\$_string\")"
		_part="${BASH_REMATCH[4]}"
		if [[ -z "${BASH_REMATCH[3]}" && -n "$_part" ]]
		then
		    [ -z "$_err_hint_vn" ] || \
			_err_hint_vn="no delimiter after DQUOTE at: $_part"
		    return 56
		fi

		# line ends with delimiter
		if [[ -n "${BASH_REMATCH[3]}" && -z "$_part" ]]
		then
		    eval "$_ret_array_vn+=(\"\")"
		    return 0
		fi
	    else
		[ -z "$_err_hint_vn" ] || \
		    _err_hint_vn="single quote followed by no delimiter: $_part"
		return 56
	    fi
	else
	    # unescaped
	    if [[ "$_part" =~ ^([^${_deli}]*)(${_deli}?)(.*)$ ]]
	    then
		eval "$_ret_array_vn+=(\"\${BASH_REMATCH[1]}\")"
		_part="${BASH_REMATCH[3]}"
		if [[ -z "${BASH_REMATCH[2]}" && -n "$_part" ]]
		then
		    bb_msg err "${FUNCNAME[0]-?}: got parse error (1) near" \
			"unescaped field which should not happen, please report"
		    [ -z "$_err_hint_vn" ] || \
			_err_hint_vn="internal error while parsing: $_part"
		    return 56
		fi

		# line ends with delimiter
		if [[ -n "${BASH_REMATCH[2]}" && -z "$_part" ]]
		then
		    eval "$_ret_array_vn+=(\"\")"
		    return 0
		fi
	    else
		bb_msg err "${FUNCNAME[0]-?}: got parse error (2) near" \
		    "unescaped field which should not happen, please report"
		[ -z "$_err_hint_vn" ] || \
		    _err_hint_vn="internal error while parsing: $_part"
		return 56
	    fi
	fi
    done
}

#@ bcsv_line_of_row <ret-vn> <fields-array-vn> <delimiter>
# creates a CSV file line out of a number of fields separated by delimiter
# errors: 1, 2, 3, 7, 11
function bcsv_line_of_row() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _fields_array_vn="$2"
    local _deli="$3"

    bb_validate_vn "${FUNCNAME[0]-?}:ret" "$_ret_vn" || return 3
    bb_validate_vn "${FUNCNAME[0]-?}:fields-array" "$_fields_array_vn" || \
	return 3

    # shellcheck disable=SC2034
    local -a bcsv_line_of_row__escaped_fields
    bb_array_map bcsv_line_of_row__escaped_fields "$_fields_array_vn" \
	bcsv_quote_field "$_deli" || {
	bb_ierr "internal error calling map: $?"
	return 1
    }

    bb_array_join_by "$_ret_vn" bcsv_line_of_row__escaped_fields "$_deli" || {
	bb_ierr "internal error calling join: $?"
	return "$?"
    }
}

#@@ Handles

#@ bcsv_is_open <handle>
# check if handle is open
# errors: 2, 51, 52
function bcsv_is_open() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _csv_handle="$1"
    bb_validate_vn bcsv_is_open "$_csv_handle" || return 51

    [[ -v "bcsv_hnd__${_csv_handle}" ]] || return 52
}

#@ bcsv_new <handle> <default-delimter>
# creates an empty CSV handle
# handle must match ^[[:alpha:]][[:alnum:]_]*$ and must be unused
# error: 2, 51, 55
function bcsv_new() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _handle="$1"
    local _deli="$2"

    bcsv_delimiter_validate "$_deli" || return "$?"

    if bcsv_is_open "$_handle"
    then
	return 55 # handle already in use
    else
	_retval="$?"
	case "$_retval" in
	    52) ;; # hande not loaded -> thats what we want
	    *) return "$_retval";; # other error
	esac
    fi

    # the following must not be done in one line
    # its a bug in bash 4.2 that is fixed in 4.4
    declare -g -A "bcsv_hnd__${_handle}"
    eval "bcsv_hnd__${_handle}=([deli]=\"\$_deli\" [rows]=0)"
}

#@ bcsv_discard <handle>
# discard ever
# errors, 2, 51, 52
function bcsv_discard() {
    (( $# == 1 )) || {
	bb_ierr "called with $# instead of 1 arg"
	return 2
    }

    local _handle="$1"

    bcsv_is_open "$_handle" || return "$?"

    local -i bcsv_discard__rows
    bcsv_row_size bcsv_discard__rows "$_handle"

    local -i _i
    for (( _i = 1; _i <= bcsv_discard__rows; _i++ ))
    do
	unset "bcsv_hnd__${_handle}_row_${_i}"
    done

    unset "bcsv_hnd__${_handle}"
}

#@@ Access and Manipulation of loaded CSVs

#@ bcsv_delimiter_get <ret-vn> <handle>
# returns configured delimiter used by handle
# errors: 2, 3, 51, 52
function bcsv_delimiter_get() {
   (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _handle="$2"

    bb_validate_vn "${FUNCNAME[0]-?}:ret-vn" "$_ret_vn" || return 3
    bcsv_is_open "$_handle" || return "$?"

    local _ref="bcsv_hnd__${_handle}[deli]"
    printf -v "$_ret_vn" "%s" "${!_ref}"
}

#@ bcsv_delimiter_set <handle> <delimiter>
# sets configured delimiter used by handle
# errors: 2, 3, 51, 52
function bcsv_delimiter_set() {
   (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _handle="$1"
    local _new_deli="$2"

    bcsv_is_open "$_handle" || return "$?"
    bcsv_delimiter_validate "$_deli" || return "$?"

    eval "bcsv_hnd__${_handle}[deli]=\"$_new_deli\""
}

#@ bcsv_row_size <ret-rows-vn> <handle>
# get number of rows
# errors: 2, 3, 51, 52
function bcsv_row_size() {
    (( $# == 2 )) || {
	bb_ierr "called with $# instead of 2 args"
	return 2
    }

    local _ret_rows_vn="$1"
    local _handle="$2"

    bb_validate_vn "bcsv_row_size:ret-rows-vn" "$_ret_rows_vn" || return 3
    bcsv_is_open "$_handle" || return "$?"

    local _ref="bcsv_hnd__${_handle}[rows]"
    printf -v "$_ret_rows_vn" "%i" "${!_ref}"
}

#@ bcsv_row_get <ret-row-array-vn> <handle> <row-number>
# errors: 2, 3, 51, 52, 54
function bcsv_row_get() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _handle="$2"
    local _row_nr="$3"

    bb_validate_uint "${FUNCNAME[0]-?}:row-number" "$_row_nr" || return 3
    (( _row_nr > 0 )) || return 3

    # checks (_ret_vn, _handle) are done by bcsv_row_size

    local -i bcsv_row_get__rows
    bcsv_row_size bcsv_row_get__rows "$_handle"

    (( _row_nr <= bcsv_row_get__rows )) || return 54

    bb_array_copy "$_ret_vn" "bcsv_hnd__${_handle}_row_${_row_nr}"
}

#@ bcsv_row_find <ret-row-number-vn> <handle> <column-number> <value>
# returns number of row which has value in column column-number
# errors: 2, 3, 10, 51, 52, 54
function bcsv_row_find() {
    (( $# == 4 )) || {
	bb_ierr "called with $# instead of 4 args"
	return 2
    }

    local _ret_vn="$1"
    local _handle="$2"
    local _col_nr="$3"
    local _value="$4"

    bb_validate_vn "${FUNCNAME[0]-?}:ret-row-number" "$_ret_vn" || return 3
    bb_validate_uint "${FUNCNAME[0]-?}:column-number" "$_col_nr" || return 3

    (( _col_nr > 0 )) || return 3
    _col_nr="$((_col_nr - 1 ))"

    # shellcheck disable=SC2034
    local -a bcsv_row_get__fields
    local -i bcsv_row_get__rows _i _retval
    bcsv_row_size bcsv_row_get__rows "$_handle"

    # shellcheck disable=SC2034
    local bcsv_row_get__value
    # shellcheck disable=SC2154
    for (( _i=1; _i <= bcsv_row_get__rows; _i++ ))
    do
	bcsv_row_get bcsv_row_get__fields "$_handle" "$_i" || return "$?"
	if bb_array_get bcsv_row_get__value bcsv_row_get__fields "$_col_nr"
	then
	    if [ "$bcsv_row_get__value" == "$_value" ]
	    then # we found what we are looking for
		printf -v "$_ret_vn" "%i" "$_i"
		return 0
	    fi
	else # if value not found, ignore, on other error return it
	    _retval="$?"
	    (( _retval == 10 )) || return "$_retval"
	fi
    done

    # we still have'nt found what we are looking for
    return 10
}

#@ bcsv_row_set <handle> <row-number> <fields-array-vn>
# set a row, overwriting any potential contents
# if row-number is empty it will behave like bcsv_row_add
# if row-number > current row size empty rows will be filled in
# errors: 2, 3, 51, 52
function bcsv_row_set() {
    (( $# == 3)) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _handle="$1"
    local _row_nr="$2"
    local _fields_vn="$3"

    bb_var_assert_array "$_fields_vn" "fields-array-vn" || return "$?"

    [ -z "$_row_nr" ] || {
	bb_validate_uint "bcsv_row_set:row-number" "$_row_nr" || return 3
	(( _row_nr > 0 )) || return 3
    }
    bcsv_is_open "$_handle" || {
	local _retval="$?"
	bb_msg warning "bcsv_row_set: handle_is_open($_handle): $_retval"
	return "$_retval"
    }

    local -i bcsv_row_set__rs _i
    bcsv_row_size bcsv_row_set__rs "$_handle"

    # if no row number is specified use next free
    [ -n "$_row_nr" ] || _row_nr="$(( bcsv_row_set__rs + 1 ))"

    # if row is outside current size, fill up with empty rows
    for (( _i = bcsv_row_set__rs + 1; _i < _row_nr; _i++ ))
    do
	eval "bcsv_hnd__${_handle}_row_${_i}=()"
    done

    declare -g -a "bcsv_hnd__${_handle}_row_${_row_nr}"
    bb_array_copy "bcsv_hnd__${_handle}_row_${_row_nr}" "$_fields_vn" || \
	return 1

    (( _row_nr < bcsv_row_set__rs )) || \
	eval "bcsv_hnd__${_handle}[rows]=$(( _row_nr ))"
}

#@ bcsv_row_set_direct <handle> <row-number> [<field> [..]]
# set a row, overwriting any potential contents
# if row-number is empty it will behave like bcsv_row_add
# if row-number > current row size empty rows will be filled in
# errors: 2, 3, 51, 52
function bcsv_row_set_direct() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    local _handle="$1"
    local _row_nr="$2"
    shift 2

    # shellcheck disable=SC2034
    local -a bcsv_row_set_direct__fields
    if (( $# > 0 ))
    then
	eval "bcsv_row_set_direct__fields=(\"\$@\")"
    else
	eval "bcsv_row_set_direct__fields=()"
    fi
    bcsv_row_set "$_handle" "$_row_nr" bcsv_row_set_direct__fields || \
	return "$?"
}

#@ bcsv_row_add <handle> <field> [..]
# adds a row
# errors: 2, 51, 52
function bcsv_row_add() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    local _handle="$1"
    shift

    bcsv_row_set_direct "$_handle" "" "$@"
}

#@@ Iterators

#@ bcsv_row_iter <handle> <iter-fun> [<xarg> [..]]
# iterates over a CSV row by row
# iter-fun must take: <fields-array-vn> <index> [<xarg> [..]]
# errors: 2, 3, 6, 8, 91 (one or more iter-fun calls failed)
function bcsv_row_iter() {
    (( $# > 1 )) || {
	bb_ierr "called with $# instead of 2 or more args"
	return 2
    }

    local _handle="$1"
    local _iter_fun="$2"
    shift 2

    bb_funs_defined bb_csv_row_iter "$_iter_fun" || return 6

    local -i bcsv_row_iter__rs _i _retval=0
    bcsv_row_size bcsv_row_iter__rs "$_handle" || return "$?"

    for (( _i=1; _i <= bcsv_row_iter__rs; _i++ ))
    do
	"$_iter_fun" "bcsv_hnd__${_handle}_row_${_i}" "$_i" "$@" || _retval="91"
    done

    return "$_retval"
}

#@@ CSV Processing

#@ bcsv_row_map <ret-handle> <in-handle> <map-fun> [<xarg>] [..]]
# applys map-fun to every row of a CSV, generating a new CSV as output
# mapfun must take: <ret-fields-array-vn> <fields-array-vn> <index>
#                   [<xarg>] [..]]
# errors: 2, 3, 6, 8, 91 (one or more map-fun calls failed)
function bcsv_row_map() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    local _out_handle="$1"
    local _in_handle="$2"
    local _map_fun="$3"
    shift 3

    bb_funs_defined bb_csv_row_map "$_map_fun" || return 6

    # shellcheck disable=SC2034
    local -i bcsv_row_map__rs _i _retval=0
    bcsv_row_size bcsv_row_map__rs "$_in_handle" || return "$?"

    local bcsv_row_map__deli
    bcsv_delimiter_get bcsv_row_map__deli "$_in_handle" || return "$?"

    bcsv_new "$_out_handle" "$bcsv_row_map__deli" || return "$?"

    local -i _retval=0
    # shellcheck disable=SC2034
    local -a bcsv_row_map__fields_src bcsv_row_map__fields_dst
    for (( _i=1; _i <= bcsv_row_map__rs; _i++ ))
    do
	bcsv_row_get bcsv_row_map__fields_src "$_in_handle" "$_i" || return "$?"
	"$_map_fun" bcsv_row_map__fields_dst bcsv_row_map__fields_src \
	    "$_i" "$@" || _retval="$?"
        bcsv_row_set "$_out_handle" "" bcsv_row_map__fields_dst || return "$?"
    done

echo "ROW_MAP done, retval=$_retval"
    return "$_retval"
}

#@ bcsv_row_filter <ret-handle> <in-handle> <filter-fun> [<xarg>] [..]]
# create a new CSV containing all rows in source CSV that run filter-fun
# successfully
# filterfun must take: <fields-array-vn> <index> [<xarg>] [..]]
# errors: 2, 3, 6, 8, 91 (one or more filter-fun calls failed)
function bcsv_row_filter() {
    (( $# > 2 )) || {
	bb_ierr "called with $# instead of 3 or more args"
	return 2
    }

    local _out_handle="$1"
    local _in_handle="$2"
    local _filter_fun="$3"
    shift 3

    bb_funs_defined bb_csv_row_filter "$_filter_fun" || return 6

    # shellcheck disable=SC2034
    local -i bcsv_row_filter__rs _i _retval=0
    bcsv_row_size bcsv_row_filter__rs "$_in_handle" || return "$?"

    local bcsv_row_filter__deli
    bcsv_delimiter_get bcsv_row_filter__deli "$_in_handle" || return "$?"

    bcsv_new "$_out_handle" "$bcsv_row_filter__deli" || return "$?"

    # shellcheck disable=SC2034
    local -a bcsv_row_filter__fields_src
    for (( _i=1; _i <= bcsv_row_filter__rs; _i++ ))
    do
	bcsv_row_get bcsv_row_filter__fields_src "$_in_handle" "$_i" || \
	    return "$?"
	if "$_filter_fun" bcsv_row_filter__fields_src "$_i" "$@"
	then
	    bcsv_row_set "$_out_handle" "" bcsv_row_filter__fields_src
	fi
    done
}

#@@ File IO

#@ bcsv_load <handle> <delimiter> (<csv-file>|-) [<ret-error-hint-vn>]
# loads contents of csv-file or from stdin if '-' is specified
# handle must match ^[[:alpha:]][[:alnum:]_]*$ and must be unused
# ret-error-hint will be filled with error hint if specified
# errors: 2, 8, 9, 51, 55
function bcsv_load() {
    (( $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    local _handle="$1"
    local _deli="$2"
    local _input_file="$3"
    local _err_hint="${4-}"

    bcsv_new "$_handle" "$_deli" || return "$?"

    local _in_path
    if [ "$_input_file" == "-" ]
    then
	_in_path="/dev/stdin"
    else
	_in_path="$_input_file"
	[ -f "$_in_path" ] || return 8
	[ -r "$_in_path" ] || return 9
    fi

    local -i _retval _line_nr=1
    local _line bcsv_load__err_hint
    local -a bcsv_load__fields

    # shellcheck disable=SC2094
    while read -r _line || [ -n "$_line" ]
    do
	_bcsv_fields_of_line bcsv_load__fields "$_deli" "$_line" \
	    bcsv_load__err_hint || {
	    _retval="$?"

	    [ -z "$_err_hint" ] || printf -v "$_err_hint" \
		"parse error reading %s line %i: %s" \
		"$_in_path" "$(( _i + 1 ))" "$_line_nr" "$bcsv_load__err_hint"

	    return "$_retval"
	}

	if (( ${#bcsv_load__fields[@]} > 0 ))
	then
	    bcsv_row_set_direct "$_handle" "" "${bcsv_load__fields[@]}" || return "$?"
	else
	    bcsv_row_set_direct "$_handle" "" || return "$?"
	fi
	_line_nr=$((_line_nr + 1))
    done < "$_in_path"
}

# _bcsv_save_iterator <fields-array-vn> <index> <deli>
# errors: 2
function _bcsv_save_iterator() {
    (( $# == 3 )) || {
	bb_ierr "called with $# instead of 3 args"
	return 2
    }

    local _fields_array="$1"
    local _index="$2"
    local _deli="$3"

    if bb_array_is_empty "$_fields_array"
    then
	# no elements, just print empty line
	printf "<>\n"
	return 0
    fi

    # shellcheck disable=SC2034
    local bcsv_save_iterator__line
    bcsv_line_of_row bcsv_save_iterator__line "$_fields_array" "$_deli" || {
	local _retval
	_retval="$?"
	bb_msg debug3 "${FUNCNAME[0]-?}: faild for row $_index"
	return "$_retval"
    }

    # shellcheck disable=SC2154
    printf "%s\n" "$bcsv_save_iterator__line"
}

#@ bcsv_save <handle> (<csv-file>|-) [<force>]
# saves CSV to file or stdout if none is specified
# if force is non empty existing file will be overwritten
# errors: 2, 9, 51, 52, 91 (file already exists)
function bcsv_save() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "called with $# instead of 2 or 3 args"
	return 2
    }

    local _handle="$1"
    local _output_file="$2"
    local _force="${3-}"

    local bcsv_save__deli

    bcsv_delimiter_get bcsv_save__deli "$_handle" || return "$?"
    bcsv_save_with_delimiter "$_handle" "$bcsv_save__deli" "$_output_file" \
	"_force"
}

#@ bcsv_save_with_delimiter <handle> <delimiter> (<csv-file>|-) [force]
# saves CSV to file or stdout if none is specified
# if force is non empty existing file will be overwritten
# uses custom delimiter specified
# errors: 2, 9, 51, 52, 91 (file already exists)
function bcsv_save_with_delimiter() {
    (( $# == 3 || $# == 4 )) || {
	bb_ierr "called with $# instead of 3 or 4 args"
	return 2
    }

    local _handle="$1"
    local _deli="$2"
    local _output_file="$3"
    local _force="${4-}"

    bcsv_delimiter_validate "$_deli" || return "$?"

    local _out_path
    if [ "$_output_file" == "-" ]
    then
	_out_path="/dev/stdin"
    else
	_out_path="$_output_file"
	if [[ -f "$_out_path" && -z "$_force" ]]
	then
	    return 91
	fi
	printf "" > "$_out_path" || return 9
    fi

    bcsv_row_iter "$_handle" _bcsv_save_iterator "$_deli" > "$_out_path" || {
	bb_msg err "${FUNCNAME[0]-?}: failed to save: $?"
	return 1
    }
}
